#include <PiPei.h>
#include <Protocol/GuidedSectionExtraction.h>
#include "Sdk/C/7zTypes.h"
#include "Sdk/C/7zVersion.h"
#include "Sdk/C/LzmaDec.h"
#include "Sdk/C/Bra.h"

#define LZMA_HEADER_SIZE (LZMA_PROPS_SIZE + 8)
#define SCRATCH_BUFFER_SIZE SIZE_64KB

typedef struct
{
  ISzAlloc Functions;
  VOID     *Buffer;
  UINTN    BufferSize;
} ISzAllocWithData;

EFI_STATUS
EFIAPI
LzmaF86GuidedSectionExtract (
  IN CONST  EFI_GUIDED_SECTION_EXTRACTION_PROTOCOL *This,
  IN CONST  VOID                                   *InputSection,
  OUT       VOID                                   **OutputBuffer,
  OUT       UINTN                                  *OutputSize,
  OUT       UINT32                                 *AuthenticationStatus
  );

EFI_HANDLE         gImageHandle = NULL;
EFI_SYSTEM_TABLE   *gST = NULL;
EFI_BOOT_SERVICES  *gBS = NULL;

EFI_GUIDED_SECTION_EXTRACTION_PROTOCOL mLzmaF86GuidedSectionExtractionProtocol = {
  LzmaF86GuidedSectionExtract
};

VOID *
EFIAPI
CopyMem (
  OUT VOID       *DestinationBuffer,
  IN CONST VOID  *SourceBuffer,
  IN UINTN       Length
  )
{
  gBS->CopyMem (DestinationBuffer, (VOID*) SourceBuffer, Length);
  return EFI_SUCCESS;
}

BOOLEAN
EFIAPI
CompareGuid (
  IN CONST GUID  *Guid1,
  IN CONST GUID  *Guid2
  )
{
  UINT64  LowPartOfGuid1;
  UINT64  LowPartOfGuid2;
  UINT64  HighPartOfGuid1;
  UINT64  HighPartOfGuid2;

  LowPartOfGuid1  = *((CONST UINT64*) Guid1);
  LowPartOfGuid2  = *((CONST UINT64*) Guid2);
  HighPartOfGuid1 = *((CONST UINT64*) Guid1 + 1);
  HighPartOfGuid2 = *((CONST UINT64*) Guid2 + 1);

  return (BOOLEAN) (LowPartOfGuid1 == LowPartOfGuid2 && HighPartOfGuid1 == HighPartOfGuid2);
}

VOID *
SzAlloc (
  VOID *P,
  size_t Size
  )
{
  VOID *Addr;
  ISzAllocWithData *Private;

  Private = (ISzAllocWithData*) P;

  if (Private->BufferSize >= Size) {
    Addr = Private->Buffer;
    Private->Buffer = (VOID*) ((UINT8*)Addr + Size);
    Private->BufferSize -= Size;
    return Addr;
  } else {
    return NULL;
  }
}

VOID
SzFree (
  VOID *P,
  VOID *Address
  )
{
  //
  // We use the 'scratch buffer' for allocations, so there is no free
  // operation required.  The scratch buffer will be freed by the caller
  // of the decompression code.
  //
}

UINT64
GetDecodedSizeOfBuf(
  UINT8 *EncodedData
  )
{
  UINT64 DecodedSize;
  INTN   Index;

  /* Parse header */
  DecodedSize = 0;
  for (Index = LZMA_PROPS_SIZE + 7; Index >= LZMA_PROPS_SIZE; Index--)
    DecodedSize = (DecodedSize << 8) + EncodedData[Index];

  return DecodedSize;
}

EFI_STATUS
EFIAPI
LzmaF86GuidedSectionExtract (
  IN CONST  EFI_GUIDED_SECTION_EXTRACTION_PROTOCOL *This,
  IN CONST  VOID                                   *InputSection,
  OUT       VOID                                   **OutputBuffer,
  OUT       UINTN                                  *OutputSize,
  OUT       UINT32                                 *AuthenticationStatus
  )
{
  EFI_STATUS       Status;
  EFI_GUID         *InputGuid;
  VOID             *Source;
  UINTN            SourceSize;
  VOID             *ScratchBuffer;
  VOID             *AllocatedOutputBuffer;
  UINTN            OutputBufferSize;
  SRes             LzmaResult;
  ELzmaStatus      LzmaStatus;
  SizeT            EncodedDataSize;
  ISzAllocWithData AllocFuncs;
  UINT32           X86State;

  if (IS_SECTION2 (InputSection)) {
    InputGuid  = &(((EFI_GUID_DEFINED_SECTION2 *) InputSection)->SectionDefinitionGuid);
    Source     = (UINT8 *) InputSection + ((EFI_GUID_DEFINED_SECTION2 *) InputSection)->DataOffset;
    SourceSize = SECTION2_SIZE (InputSection) - ((EFI_GUID_DEFINED_SECTION2 *) InputSection)->DataOffset;
  } else {
    InputGuid  = &(((EFI_GUID_DEFINED_SECTION *) InputSection)->SectionDefinitionGuid);
    Source     = (UINT8 *) InputSection + ((EFI_GUID_DEFINED_SECTION *) InputSection)->DataOffset;
    SourceSize = SECTION_SIZE (InputSection) - ((EFI_GUID_DEFINED_SECTION *) InputSection)->DataOffset;
  }

  if (!CompareGuid (&gLzmaF86CustomDecompressGuid, InputGuid)) {
    return RETURN_INVALID_PARAMETER;
  }

  if (SourceSize < LZMA_HEADER_SIZE) {
    return RETURN_INVALID_PARAMETER;
  }

  //
  // Call GetInfo to get the size and attribute of input guided section data.
  //
  OutputBufferSize = (UINTN) GetDecodedSizeOfBuf ((UINT8*) Source);

  if (OutputBufferSize > 0) {
    //
    // Allocate output buffer
    //
    Status = gBS->AllocatePool (EfiBootServicesData, OutputBufferSize, &AllocatedOutputBuffer);
    if (EFI_ERROR (Status)) {
      return Status;
    }

    //
    // Allocate scratch buffer
    //
    Status = gBS->AllocatePool (EfiBootServicesData, SCRATCH_BUFFER_SIZE, &ScratchBuffer);
    if (EFI_ERROR (Status)) {
      goto FreeAllocatedOutputBuffer;
    }

    //
    // Call decode function to extract raw data from the guided section.
    //
    AllocFuncs.Functions.Alloc  = SzAlloc;
    AllocFuncs.Functions.Free   = SzFree;
    AllocFuncs.Buffer           = ScratchBuffer;
    AllocFuncs.BufferSize       = SCRATCH_BUFFER_SIZE;
  
    EncodedDataSize = (SizeT) (SourceSize - LZMA_HEADER_SIZE);

    LzmaResult = LzmaDecode(
      AllocatedOutputBuffer,
      (SizeT*) &OutputBufferSize,
      (Byte*) ((UINT8*) Source + LZMA_HEADER_SIZE),
      &EncodedDataSize,
      Source,
      LZMA_PROPS_SIZE,
      LZMA_FINISH_END,
      &LzmaStatus,
      &(AllocFuncs.Functions)
      );

    //
    // Free unused scratch buffer.
    //
    gBS->FreePool (ScratchBuffer);

    if (LzmaResult != SZ_OK) {
      //
      // Decode failed
      //
      Status = RETURN_INVALID_PARAMETER;
      goto FreeAllocatedOutputBuffer;
    }

    //
    // After decompress, the data need to be converted to the raw data. 
    //
    x86_Convert_Init (X86State);
    x86_Convert (AllocatedOutputBuffer, OutputBufferSize, 0, &X86State, 0);
  } else {
    AllocatedOutputBuffer = NULL;
  }

  *OutputBuffer = AllocatedOutputBuffer;

  //
  // Set real size of output buffer.
  //
  *OutputSize = OutputBufferSize;

  //
  // Authentication is set to Zero, which may be ignored.
  //
  *AuthenticationStatus = 0;

  return EFI_SUCCESS;

FreeAllocatedOutputBuffer:
  gBS->FreePool(AllocatedOutputBuffer);
  return Status;
}

EFI_STATUS
EFIAPI
_ModuleEntryPoint(
  IN EFI_HANDLE                   ImageHandle,
  IN EFI_SYSTEM_TABLE             *SystemTable
  )
{
  EFI_STATUS                      Status;
  VOID                            *Interface;
  EFI_HANDLE                      Handle;
  VOID                            *GuidData;

  gImageHandle = ImageHandle;
  gST = SystemTable;
  gBS = SystemTable->BootServices;

  Status = gBS->LocateProtocol(&gLzmaF86CustomDecompressGuid, NULL, &Interface);
  if (!EFI_ERROR (Status)) {
    return EFI_INVALID_PARAMETER;
  }

  Handle = NULL;
  Status = gBS->InstallProtocolInterface (&Handle, &gLzmaF86CustomDecompressGuid, EFI_NATIVE_INTERFACE, &mLzmaF86GuidedSectionExtractionProtocol);
  if (EFI_ERROR (Status)) {
    return Status;
  }

  Status = gBS->AllocatePool (EfiBootServicesData, sizeof (GUID), &GuidData);
  if (EFI_ERROR (Status)) {
    goto UninstallProtocolInterface;
  }

  gBS->CopyMem (GuidData, &gLzmaF86CustomDecompressGuid, sizeof(GUID));

  Status = gBS->InstallConfigurationTable (&gLzmaF86CustomDecompressGuid, GuidData);
  if (EFI_ERROR (Status)) {
    goto FreeGuidData;
  }

  return EFI_SUCCESS;

FreeGuidData:
  gBS->FreePool (GuidData);

UninstallProtocolInterface:
  gBS->UninstallProtocolInterface (Handle, &gLzmaF86CustomDecompressGuid, &mLzmaF86GuidedSectionExtractionProtocol);
  return Status;
}
